################################################################
# Printer package for Unix
# Emulate printer package for NT in a "reasonable" way
# $Header: C:/CHINOOK/NEWPROD/TCL/printer/RCS/printunx.tcl 1.11 2004/08/15 16:12:16 Michael_Schwartz Exp Michael_Schwartz $
# RCS Revision History
# $Log: printunx.tcl $
# Revision 1.11  2004/08/15 16:12:16  Michael_Schwartz
# Updated for Tk 8.4, which moved the FocusGroup commands to "unsupported"
#
# Revision 1.10  2002/10/08 17:50:23  Michael_Schwartz
# Changed use of profile package to new name of ini
#
# Revision 1.9  2000/07/03 08:31:36  Michael_Schwartz
# Updated use of namespaces to full qualification.
# Updated lpstat parsing to look for colons on otherwise non-blank lines
#
# Revision 1.8  2000/07/02 23:35:07  Michael_Schwartz
# Added small fix and acknowledgement
#
# Revision 1.7  1999/09/09 04:56:58  Michael_Schwartz
# *** empty log message ***
#
# Revision 1.6  1998/05/03  16:49:42  Michael_Schwartz
# Initial tests with OSF/1 complete successfully
#
# Revision 1.5  1998/04/19  18:40:03  Michael_Schwartz
# Added selection and setup dialogs.
#
# Revision 1.4  1998/04/16  05:11:34  Michael_Schwartz
# Better tracking for list versus entry; removed unused code
#
# Revision 1.3  1998/04/16  04:22:54  Michael_Schwartz
# Combined with tkPBox.tcl.
# Updated to make nominal cases succeed.
#
# Revision 1.2  1998/04/14  22:43:35  Michael_Schwartz
# Added initial bodies for most commands.
#
################################################################
## Copyright (C) Schwartz Computer Consulting Services, 1998
##
## {LICENSE}
## 
## THE AUTHORS HEREBY GRANT PERMISSION TO USE, COPY, MODIFY, DISTRIBUTE,
## AND LICENSE THIS SOFTWARE AND ITS DOCUMENTATION FOR ANY PURPOSE, PROVIDED
## THAT EXISTING COPYRIGHT NOTICES ARE RETAINED IN ALL COPIES AND THAT THIS
## NOTICE IS INCLUDED VERBATIM IN ANY DISTRIBUTIONS. 
##
## NO WRITTEN AGREEMENT, LICENSE, OR ROYALTY FEE IS REQUIRED FOR ANY OF THE
## AUTHORIZED USES.
## 
## MODIFICATIONS TO THIS SOFTWARE MAY BE COPYRIGHTED BY THEIR AUTHORS
## AND NEED NOT FOLLOW THE LICENSING TERMS DESCRIBED HERE, PROVIDED THAT
## THE NEW TERMS ARE CLEARLY INDICATED ON THE FIRST PAGE OF EACH FILE WHERE
## THEY APPLY.
## 
## IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT,
## INDIRECT, SPECIAL, INCIDENTAL,  OR CONSEQUENTIAL DAMAGES ARISING OUT OF
## THE USE OF THIS SOFTWARE, ITS DOCUMENTATION,  OR ANY DERIVATIVES
## THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE POSSIBILITY OF
## SUCH DAMAGE.
## 
## THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
## INCLUDING, BUT NOT LIMITED TO,  THE IMPLIED WARRANTIES OF
## MERCHANTABILITY,FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT. 
## THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND
## DISTRIBUTORS HAVE NO OBLIGATION  TO PROVIDE MAINTENANCE, SUPPORT,
## UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
## 
## Thanks to Kazunori Higuchi <higuchi@cov.tytlabs.co.jp> for reporting
## a bug in the Unix printer package, and providing a fix.
################################################################

# Need to initiate the namespace or it doesn't exist
namespace eval printer { }


# Global variables related to the printer
proc printer::init_globals { } {
  variable ::printer::channel         ;# Global channel for multiple pipes to printer
  variable ::printer::lpcmd           ;# Print command name
  variable ::printer::lpcmd_devarg    ;# Flag for printer command to specify a device
  variable ::printer::lpcmd_device    ;# Value for the printer queue device (name from lpstat)
  variable ::printer::lplistcmd       ;# Basis of printer list command
  variable ::printer::commandlist     ;# List of allowable subcommands for printer 
  variable ::printer::usage           ;# Standard usage message
  variable ::printer::lpcmd_extra     ;# Additional arguments to the lp command
  variable ::printer::lplist_usecache ;# Flag (non-zero) to preserve list of printers

  # It's possible that the actual locations of lp and lpstat may differ by system
  # e.g., /bin vs. /usr/bin.
  # In that case, the plan is to create a table by OS type or version, and make
  # the set as a result of a switch statement.
  # Storing the results in a profile table will work as well
  set printer::channel          ""
  set printer::lpcmd   		/bin/lp
  set printer::lpcmd_devarg     -d
  set printer::lpcmd_device     ""
  set printer::lplistcmd	"/bin/lpstat -a"
  set printer::commandlist      {attr close dialog job list open page send version}
  set printer::lpcmd_extra      ""
  set printer::lplist_usecache 0

  set printer::usage { Usage:
    printer attr
    printer close
    printer dialog [select|page_setup] [-flags flagsnum]
    printer job [start|end]
    printer list [-match matchstring]
    printer open
    printer page [start|end]
    printer send [-postscript|-nopostscript] [-hDC hdc] [-printer pname]
                 [-file|-data] file_or_data ...
    printer version
  }
  # In the more recent versions of Tk, the focus group commands are not recommended.
  # These are used to ensure the dialog becomes proper modality; please replace with the 
  # correct modern equivalent when determined. Used in tkPDialog_Create.
  # If you are using an earlier version of tk, these commands may be omitted
  global tk_version
  if { "$tk_version" >= "8.4" } {
    ::tk::unsupported::ExposePrivateCommand tkFocusGroup_Create
    ::tk::unsupported::ExposePrivateCommand tkFocusGroup_BindOut
  }
}
################################################################
## printer
## This procedure looks up the subcommand, returning the usage
## message if unsuccessful.
################################################################
proc printer { args } {
  variable ::printer::usage

  set cmd [lindex $args 0]
  if { [lsearch $printer::commandlist $cmd] == -1 } {
    error $printer::usage
  }
  eval printer::$cmd [lrange $args 1 end]  
}

################################################################
## attr
## This function returns the current set of printer attributes
## This should model the NT version as closely as possible.
## Thus, the attributes to be considered are:
## copies, first page, last page, hDC, device, page orientation,
## resolution, pixels per inch, page dimensions, page margins,
## page minimum margins 
################################################################
proc printer::attr { args } {
  variable ::printer::channel
  variable ::printer::lpcmd_device

  if { $printer::channel == "" } {
    set hDC ?
  } else {
    set hDC $printer::channel
  }

  if { $printer::lpcmd_device == "" } {
    set dev default
  } else {
    set dev $printer::lpcmd_device
  }

  lappend retval [::list copies ?]
  lappend retval [::list "first page" ?]
  lappend retval [::list "last page"  ?]
  lappend retval [::list "hDC" $hDC]
  lappend retval [::list device $dev]
  lappend retval [::list "page orientation"  ?]
  lappend retval [::list "resolution"  ?]
  lappend retval [::list "pixels per inch"  ?]
  lappend retval [::list "page dimensions"  "? ?"]
  lappend retval [::list "page margins"  "? ? ? ?"]
  lappend retval [::list "minimum page margins"  "? ? ? ?"]

  return $retval
}

################################################################
## close
## This procedure closes any open printer channel and blanks
## out the channel variable
################################################################
proc printer::close { args } {
  # This command closes and invalidates the global channel
  variable ::printer::channel
  if { [string length $printer::channel] > 0 } {
    ::close $printer::channel
    set printer::channel ""
  }
}

################################################################
## dialog
## Bring up either the printer selection dialog or the page_setup
## dialog
## For Unix, these provide the way to set the printer and 
## provide any other arbitrary arguments to the print command.
################################################################
proc printer::dialog { args } {
  variable ::printer::lpcmd                 ;# Print command name
  variable ::printer::lpcmd_devarg          ;# Flag for printer command to specify a device
  variable ::printer::lplistcmd             ;# Basis of printer list command
  variable ::printer::lpcmd_extra	          ;# Additional arguments to the lp command
  variable ::printer::usage		  ;# Usage error message
  # [select|page_setup] [-flags flagsnum] 
  variable ::printer::lpcmd_device	  ;# Value for the printer queue device (name from lpstat)

  # To be equivalent to the NT printing model, these functions should
  # actually open the channel to the selected printer.
  # However, for now, as the "generic" printer model is still under thought,
  # I'll be a bit schizoid and let this version just return the selected printer
  # name. It will set the "global" default for this printer for all future "opens",
  # though.
  switch [lindex $args 0] {
    select	{
		  # Build a single-selection list; highlight current device
		  set result [ tkPDialog -printers [printer::list] ]
		  if { [lindex $result 0] != 0 } {
		    set printer::lpcmd_device [lindex $result 1]
		    if { [lindex $result 0] == 2 } {
		      if [ catch { package require ini } err ] {
                        ;# Do nothing if we can't set global values...
                      } else {
			ini set -section Printer -key Default -value $printer::lpcmd_device
                      }
                    }
		  }
		  return $result	;# Actually, should return an "hDC" equivalent
		}
    page_setup	{
		  # Build a single entry widget; initialize with lpcmd_extra
		  set result [ tkPOption -options $printer::lpcmd_extra ]
		  if { [lindex $result 0] != 0 } {
                    set printer::lpcmd_extra [lindex $result 1]
		    if { [lindex $result 0] == 2 } {
		      if [ catch { package require ini } err ] {
                        ;# Do nothing if we can't set global values...
                      } else {
			ini set -section Printer -key Extra -value $printer::lpcmd_extra
                      }
                    }
                  }
                  return $result
		}
    default	{
		  error $printer::usage
		}
  }
}

################################################################
## job
## Jobs are started or ended
## Start will open/reopen the printer channel, using the previous
## printer command.
## End will close the printer channel.
################################################################
proc printer::job { args } {
  # [start|end] 
  # Under NT, it initializes/closes a job.
  # The best equivalent under Unix is to close and reopen a print job for start,
  # and to close and flush a print job for end.
  variable ::printer::lpcmd_last
  variable ::printer::usage
  variable ::printer::jobstate
  variable ::printer::channel
  variable ::printer::debug

  switch [lindex $args 0 ] {
    start {
            if { [info exist printer::lpcmd_last] == 0 } {
              error "Can't start print job until printer is open\n$printer::usage"
            }
	    if { [info exist printer::jobstate] && $printer::jobstate == 1 } {
	      # Restart a new job after closing the last one
              if { $printer::channel != "" } {
	        ::close $printer::channel
	      }
              # To avoid actually printing reams of paper while testing...
	      if [ info exist printer::debug ] {
	        puts "I would open $printer::lpcmd_last for write"
	      } else {
                set printer::channel [::open "|$printer::lpcmd_last" w]
	      }
            }
            
            set printer::jobstate 1
          }
    end   {
            if { [info exist printer::lpcmd_last] == 0 } {
              error "Can't stop print job until printer is open\n$printer::usage"
            }
	    if { [info exist printer::jobstate] == 0 || $printer::jobstate != 1 } {
	      error "Can't stop print job until it is started\n$printer::usage"
            }
            set printer::jobstate 0
	    if { [string length $printer::channel] > 0 } {
              ::close $printer::channel
	    }
            set $printer::channel ""
          }
    default {
	    error "printer job [start|stop]\n$printer::usage"
          }
  }
}

################################################################
## list
## Returns the list of printers; this is the same list that 
## is displayed in the printer dialog select command.
## Like the Windows version, the match currently is not supported
################################################################
proc printer::list { args } {
  # [-match matchstring] 
  # This one is difficult only because lpstat and lpq have such different results.
  # For now, build to lpstat model--1st arg on each line is the item
  # If we get fancy, the 2nd arg should be "accepting"
  variable ::printer::lplistcmd         ;# Basis of printer list command
  variable ::printer::lplist
  variable ::printer::lplist_usecache
  variable ::printer::lpcmd_device

  if { [info exist ::printer::lplist] && [info exist ::printer::lplist_usecache] } {
    if { $::printer::lplist_usecache != 0 } {
      return $::printer::lplist
    }
  }

  set ::printer::lplist ""
  catch {
    set fd [ ::open "|$::printer::lplistcmd" r ]

    while { [ gets $fd line ] > 0 } {
      # If the last character is a colon, this is probably an lpq-type command
      if { [string range $line 0 0] != " " && \
           [string range $line 0 0] != "\t" && \
           [string range $line end end] == ":" } {
        # Everything is cool if we skip the colon
        lappend ::printer::lplist [ string trimright $line :]
      } elseif { [string range $line 0 0] == " " || [ string range $line 0 0 ] == "\t" } {
        ;# It's an lpq-type command--ignore these lines
      } else {
        # Otherwise, it's pretty normal--just use the first word
        lappend ::printer::lplist [lindex $line 0]
      }
    }

    ::close $fd
  } err

  # If a specific device has already been selected, allow it to be
  # selected again.
  if [ info exist ::printer::lpcmd_device ] {
    if { [lsearch -exact $::printer::lplist $::printer::lpcmd_device] == -1 } {
      lappend ::printer::lplist $::printer::lpcmd_device
    }
  }

  return $::printer::lplist   
}

################################################################
## open
## This command will open the currently specified printer using
## the configured command, or open the default printer if nothing
## else is specified.
################################################################
proc printer::open { args } {
  # Should open a file output to the selected printer & return the file handle
  # For now, assume the first argument is a printer name, and ignore all others.
  # The caller will have to handle any exceptions raised, since no policy can
  # be established by this routine.
  variable ::printer::lpcmd
  variable ::printer::lpcmd_devarg
  variable ::printer::lpcmd_device	;# Value for the printer queue device (name from lpstat)
  variable ::printer::lpcmd_last
  variable ::printer::channel
  variable ::printer::debug

  set cmd $::printer::lpcmd

  if { [llength $args] > 0 } {
    set cmd "$cmd $::printer::lpcmd_devarg[lindex $args 0]"
  } elseif [info exist ::printer::lpcmd_device] {
    set cmd "$cmd $::printer::lpcmd_devarg$::printer::lpcmd_device"
  }

  set ::printer::lpcmd_last $cmd
  # To test without printing, the following debug is added
  if [ info exist ::printer::debug ] {
    puts "I would open $cmd for write"
  } else {
    set ::printer::channel [::open "|$cmd" w]
  }
  return $::printer::channel
}

################################################################
## page
## Provided for compatibility; to do anything useful, at a minimum
## the type of file currently printed would need to be known.
## Currently ignored.
################################################################
proc printer::page { args } {
  # Not sure how to handle this. For postscript and text printers it's easy, though.
  # [start|end] 
  # For now, just ignore it.
}


################################################################
## Copy a file to a given open file descriptor.
################################################################
proc printer::copyfile { src destpipe } {
  set file [ ::open $src r ]
  fcopy $file $destpipe
  ::close $file
  flush $destpipe
}

################################################################
## send
# Lots of possibilities for send.
# It can get data or files to send to the printer
#   Files must be opened and data copied to the printer (or name sent to lp command line)
#   Data must be written to the printer standard input.
# It can open the files in binary or text mode (irrelevant in Unix(?))
# It can send to a named printer or to an open pipe.
#   For a named printer, it must open its own connection and close when done.
#   For an open pipe, it must NOT open its own connection and NOT close when done.
#    printer send [-postscript|-nopostscript] [-hDC hdc] [-printer pname]
#                 [-file|-data] file_or_data ...
#
# Strategy: Put the output in a temporary file and then send it all at once.
# This will best emulate what's happening under Windows
################################################################
proc printer::send { args } {

  variable ::printer::lpcmd         ;# Print command name
  variable ::printer::lpcmd_devarg  ;# Flag for printer command to specify a device
  variable ::printer::lplistcmd     ;# Basis of printer list command
  variable ::printer::lpcmd_device  ;# Default printer
  variable ::printer::lpcmd_extra	  ;# Additional arguments
  variable ::printer::channel

  set tmpfile "/tmp/Printer[expr int(rand() * 10000)]"
  set execprog 1
  set pipe ""
  set postscript 1
  set source file
  set nxtptr 0
  set echopart ""
  set printerpart ""
  set lppart $::printer::lpcmd
  set files ""
  set datacount 0
  set filecount 0
  set status {}

  if [ info exist ::printer::channel ] {
    if [ string length $::printer::channel ] {
      set outpipe $::printer::channel
    }
  }
  
  if [info exist ::printer::lpcmd_device] {
    set printerpart "$::printer::lpcmd_devarg$::printer::lpcmd_device"
  }
  if [info exist ::printer::lpcmd_extra] {
    set printerpart "$printerpart $::printer::lpcmd_extra"
  }

  if [ catch { ::open $tmpfile w } outfile ] {
    error "Can't open $tmpfile. $outfile. Printing aborted"
  }

  # Parse arguments as follows (should move to tclParseConfigSpec?):
  # Ignore -postscript, -nopostscript.
  # Use -printer to temporarily override the device part of the lp command
  # Use -hDC to supply an existing channel (?). Not supported yet.
  # Use -file to take future (non-flag) arguments as file names
  # Use -data to take future (non-flag) arguments as output data
  foreach name $args {
    if { $nxtptr == 0 } {
      switch -exact -- $name {
	  -postscript	{ set postscript 1 }
	  -nopostscript	{ set postscript 0 }
	  -printer	{ set nxtptr 1 }
          -hDC          { set nxtptr 2 }
	  -file		{ set source file }
	  -data		{ set source data }  
	  default	{
	  		  if { $source == "data" } {
                            puts -nonewline $outfile $name
			    incr datacount
	  		  } else {
			    if [ catch { ::open $name r } infile ] {
			      lappend status $infile
			    } else {
			      fcopy $infile $outfile
			      ::close $infile
	  		      incr filecount
			    }
	  		  }
	  		}
      }
    } elseif { $nxtptr == 1 } {
      set nxtptr 0
      set printerpart "$::printer::lpcmd_devarg$name"
      if [info exist ::printer::lpcmd_extra] {
        set printerpart "$printerpart $::printer::lpcmd_extra"
      }
    } elseif { $nxtptr == 2 } {
      set nxtptr 0
      set outpipe $name
    }
  }

  # Since the print command will produce output, that output will be returned.
  # Just in case the output is redirected to stderr (I haven't checked yet),
  # I'll 'catch' the return
  ::close $outfile

  if [ catch {
    if [info exist outpipe] {
      set infile [ ::open $tmpfile r ]
      fcopy $infile $outpipe
      ::close $infile
      flush $outpipe
    } else {
      eval exec $lppart $printerpart $tmpfile
    }
  } retval ] {
    lappend status $retval
  } else {
    # If everything worked fine, remove the temporary file
    file delete $tmpfile
  }

  lappend status "Printed $filecount files and $datacount data arguments"
  return [join $status \n]
}

################################################################
## version
################################################################
proc printer::version { args } {
  return "0.7.1.3"
}

################################################################
## read_init
## Get the default values for initializing the package.
# Defaults can be overridden globally by use of the "ini" package,
# and locally by use of $HOME/.printer.tcl file
################################################################
proc printer::read_init { } {
  # Reference the globals here for ease of use
  variable ::printer::lpcmd         ;# Print command name
  variable ::printer::lpcmd_devarg  ;# Flag for printer command to specify a device
  variable ::printer::lplistcmd     ;# Basis of printer list command
  variable ::printer::lpcmd_device  ;# Default printer
  variable ::printer::lpcmd_extra	  ;# Extra arguments (e.g. -oduplex)
  global env

  if [ catch { package require ini } err ] {
    # Do nothing. Can't set global changes
  } else {
    set dev [ini get -section Printer -key Default -default None ]
    if { $dev != "None" } {
      set ::printer::lpcmd_device $dev
    }
    set ::printer::lpcmd [ini get -section Printer -key lp -default $::printer::lpcmd]
    set ::printer::lplistcmd \
    	[ini get -section printer -key lplist -default $::printer::lplistcmd]
    set ::printer::lpcmd_devarg \
    	[ini get -section printer -key dflag -default $::printer::lpcmd_devarg]
    set arg [ini get -section Printer -key Extra -default None ]
    if { $arg != "None" } {
      set ::printer::lpcmd_extra $arg
    }
  }

  if [ file exists $env(HOME)/.printer.tcl ] {
    source $env(HOME)/.printer.tcl
  }
}

package provide printer [printer::version]

printer::init_globals

printer::read_init

################################################################
################################################################
# The following file should be in a separate file, but is
# joined with this one for conciseness and completeness in transmission.
#
# The file might be called "tkPBox.tcl", loosely based on 
# the model provided in tkFBox.tcl, used for the Unix file
# browser command.
# By comparison, this set of routines is much simpler, and
# may still have some error cases that are not handled.
#
# The user-invoked routine is 
################################################################
################################################################

# tkPBox.tcl --
#
#	Implements a "TK" standard printer selection dialog box.
#
#	The "TK" printer selection dialog box is similar to the
#	printer selection dialog box on Win95(TM). 
#
# $Header: C:/CHINOOK/NEWPROD/TCL/printer/RCS/printunx.tcl 1.11 2004/08/15 16:12:16 Michael_Schwartz Exp Michael_Schwartz $
#


#----------------------------------------------------------------------
#
#		    P R I N T E R   D I A L O G
#
#----------------------------------------------------------------------

# tkPDialog --
#
#	Implements a TK printer selection dialog.
#
# Arguments:
#	args		Options parsed by the procedure.
#	  -parent
#

proc tkPDialog {args} {
    global tkPriv
    set dataName __tk_printdialog
    upvar #0 $dataName data

    tkPDialog_Config $dataName $args

    if {![string compare $data(-parent) .]} {
        set w .$dataName
    } else {
        set w $data(-parent).$dataName
    }

    # (re)create the dialog box if necessary
    #
    if {![winfo exists $w]} {
	tkPDialog_Create $w
    } elseif {[string compare [winfo class $w] TkPDialog]} {
	destroy $w
	tkPDialog_Create $w
    }
    wm transient $w $data(-parent)

    # Withdraw the window, then update all the geometry information
    # so we know how big it wants to be, then center the window in the
    # display and de-iconify it.

    wm withdraw $w
    update idletasks
    set x [expr {[winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
	    - [winfo vrootx [winfo parent $w]]}]
    set y [expr {[winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
	    - [winfo vrooty [winfo parent $w]]}]
    wm geom $w [winfo reqwidth $w]x[winfo reqheight $w]+$x+$y
    wm deiconify $w
    wm title $w $data(-title)

    # Set a grab and claim the focus too.

    set oldFocus [focus]
    set oldGrab [grab current $w]
    if {$oldGrab != ""} {
	set grabStatus [grab status $oldGrab]
    }
    grab $w
    focus $data(ent)
    $data(ent) select from 0
    $data(ent) select to   end
    $data(ent) icursor end

    # Wait for the user to respond, then restore the focus and
    # return the index of the selected button.  Restore the focus
    # before deleting the window, since otherwise the window manager
    # may take the focus away so we can't redirect it.  Finally,
    # restore any grab that was in effect.

    tkwait variable ${dataName}(selectedPrinter)
    catch {focus $oldFocus}
    grab release $w
    wm withdraw $w
    if {$oldGrab != ""} {
	if {$grabStatus == "global"} {
	    grab -global $oldGrab
	} else {
	    grab $oldGrab
	}
    }

    return [::list $tkPriv(selectPrinter_status) $tkPriv(selectPrinter)]
}

# tkPDialog_Config --
#
#	Configures the TK filedialog according to the argument list
#
proc tkPDialog_Config {dataName argList} {
    upvar #0 $dataName data

    # 1: the configuration specs
    #

    set specs {
	{-parent "" "" "."}
	{-printers "" "" "default"}
	{-title "" "" "Printer Selection"}
	{-withsave "" "" 1}
    }

    # 2: default values 
    #
    if {![info exists tkPriv(selectPrinter)]} {
	# first time the dialog has been popped up
	set tkPriv(selectPrinter) "default"
    }

    # 3: parse the arguments
    #
    tclParseConfigSpec $dataName $specs "" $argList

    if {![string compare $data(-title) ""]} {
	set data(-title) "Select Printer"
    }

    if {![winfo exists $data(-parent)]} {
	error "bad window path name \"$data(-parent)\""
    }
}

proc tkPDialog_Create {w} {
    set dataName [lindex [split $w .] end]
    upvar #0 $dataName data
    global tk_library
    global tkPriv

    toplevel $w -class TkPDialog

    # f2: the frame with the the "printer name" field
    #
    set f2 [frame $w.f2 -bd 0]
    label $f2.lab -text "Printer name:" -anchor w -width 12 -under 8 -pady 0 -padx 5
    set data(ent) [entry $f2.ent -textvariable tkPriv(selectPrinter)]

    # f1: The listbox with the printer names
    set f1 [frame $w.f1 -bd 0]
    listbox $f1.list -selectmode single -yscrollcommand "$f1.vertscroll set"
    scrollbar $f1.vertscroll -command "$f1.list yview"
    set data(list) $f1.list
    pack $f1.vertscroll -side right -fill y
    pack $f1.list -side left -fill both -expand true
    $f1.list delete 0 end

    if [ info exist data(-printers) ] {
      if { [llength $data(-printers)] > 0 } { 
        set printerlist $data(-printers)
      } else {
        set printerlist "default"
      }
    } else {
      set printerlist "default"
    }
    foreach printer $printerlist {
      $f1.list insert end $printer
    }

    # f3: the frame with the OK and cancel buttons
    #
    set f3 [frame $w.f3 -bd 0]


    set data(okBtn)     [button $f3.ok     -text OK     -under 0 -width 6 \
	-default active -pady 3]
    set data(cancelBtn) [button $f3.cancel -text Cancel -under 0 -width 6\
	-default normal -pady 3]
    if { $data(-withsave) != 0 } {
      set data(saveBtn) [button $f3.save -text Save -under 0 -width 6\
        -default normal -pady 3 -command "tkPDialog_SaveCmd $w"]
      bind $w <Alt-s> "tkButtonInvoke $data(saveBtn)"
    }

    # pack the widgets in f2 and f3
    #
    pack $data(okBtn) -side left -padx 4 -anchor center
    pack $f2.lab -side left -padx 4
    pack $f2.ent -expand yes -fill x -padx 4 -pady 0
    
    pack $data(cancelBtn) -side right -padx 4 -anchor center
    if { $data(-withsave) != 0 } {
      pack $data(saveBtn) -side left -padx 4 -anchor center -expand 1
    }

    # Pack all the frames together. We are done with widget construction.
    #
    pack $f3 -side bottom -fill x
    pack $f2 -side bottom -fill x
    pack $f1 -fill both -padx 4 -pady 1 -expand true

    # Set up the event handlers
    #
    bind $data(ent) <Return>  "tkPDialog_OkCmd $w"
    
    $data(okBtn)     config -command "tkPDialog_OkCmd $w"
    $data(cancelBtn) config -command "tkPDialog_CancelCmd $w"

    bind $w <Alt-n> "focus $data(ent)"
    bind $w <KeyPress-Escape> "tkButtonInvoke $data(cancelBtn)"
    bind $w <Alt-c> "tkButtonInvoke $data(cancelBtn)"
    bind $w.f1.list <ButtonRelease-1> { set tkPriv(selectPrinter) [%W get [ %W curselection ] ] } 
    bind $w.f1.list <KeyRelease-space>  { set tkPriv(selectPrinter) [%W get [ %W curselection ] ] } 

    wm protocol $w WM_DELETE_WINDOW "tkPDialog_CancelCmd $w"

    # Build the focus group for all the entries
    #
    tkFocusGroup_Create $w
    tkFocusGroup_BindOut $w $data(ent) "tkPDialog_EntFocusOut $w"
}

# Gets called when the entry box loses keyboard focus. Sets selection to best match 
# value in the entry box.

proc tkPDialog_EntFocusOut {w} {
    upvar #0 [winfo name $w] data
    set text [$data(ent) get]
    set list [ $data(list) get 0 end ]
    set ind [ lsearch -glob $list ${text}* ]
    if { $ind != -1 } {
      $data(list) selection set $ind
      $data(list) activate $ind
      $data(list) see $ind
    }
}

# Gets called when user presses the "OK" button
#
proc tkPDialog_OkCmd {w} {
    global tkPriv
    upvar #0 [winfo name $w] data

    set text $tkPriv(selectPrinter)
    set data(selectedPrinter) $text
    set tkPriv(selectPrinter_status) 1
    return $tkPriv(selectPrinter)
}

# Gets called when user presses the "OK" button
#
proc tkPDialog_SaveCmd {w} {
    global tkPriv
    upvar #0 [winfo name $w] data

    set text $tkPriv(selectPrinter)
    set data(selectedPrinter) $text
    set tkPriv(selectPrinter_status) 2
    return $tkPriv(selectPrinter)
}

# Gets called when user presses the "Cancel" button
#
proc tkPDialog_CancelCmd {w} {
    upvar #0 [winfo name $w] data
    global tkPriv

    set data(selectedPrinter) ""
    set tkPriv(selectPrinter_status) 0 
    return $tkPriv(selectPrinter)
}

# tkPDialog_Done --
#
#	Gets called when user has input a valid filename.  Pops up a
#	dialog box to confirm selection when necessary. Sets the
#	tkPriv(selectPrinter) variable, which will break the "tkwait"
#	loop in tkPDialog and return the selected filename to the
#	script that calls tk_getOpenFile or tk_getSaveFile
#
proc tkPDialog_Done {w {selectPrinter ""}} {
    upvar #0 [winfo name $w] data
    global tkPriv

    set data(selectedPrinter) "$selectPrinter"
    set tkPriv(selectPrinter_status) 1
    return $tkPriv(selectPrinter)
}

################################################################
################################################################
# Another file by itself--tkPOpt.tcl
################################################################
################################################################
# tkPOpt.tcl --
#
#	Implements a "TK" standard printer selection dialog box.
#
#	The "TK" printer selection dialog box is similar to the
#	printer selection dialog box on Win95(TM). 
#
# tkPOpt.tcl
# $Header: C:/CHINOOK/NEWPROD/TCL/printer/RCS/printunx.tcl 1.11 2004/08/15 16:12:16 Michael_Schwartz Exp Michael_Schwartz $
#


#----------------------------------------------------------------------
#
#		      P R I N T E R   O P T I O N   D I A L O G
#
#----------------------------------------------------------------------

# tkPOption --
#
#	Implements a TK printer option dialog.
#
# Arguments:
#	args		Options parsed by the procedure.
#	  -parent
#

proc tkPOption {args} {
    global tkPriv
    set dataName __tk_printoption
    upvar #0 $dataName data

    tkPOption_Config $dataName $args

    if {![string compare $data(-parent) .]} {
        set w .$dataName
    } else {
        set w $data(-parent).$dataName
    }

    # (re)create the dialog box if necessary
    #
    if {![winfo exists $w]} {
	tkPOption_Create $w
    } elseif {[string compare [winfo class $w] TkPDialog]} {
	destroy $w
	tkPOption_Create $w
    }
    wm transient $w $data(-parent)

    # Withdraw the window, then update all the geometry information
    # so we know how big it wants to be, then center the window in the
    # display and de-iconify it.

    wm withdraw $w
    update idletasks
    set x [expr {[winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
	    - [winfo vrootx [winfo parent $w]]}]
    set y [expr {[winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
	    - [winfo vrooty [winfo parent $w]]}]
    wm geom $w [winfo reqwidth $w]x[winfo reqheight $w]+$x+$y
    wm deiconify $w
    wm title $w $data(-title)

    # Set a grab and claim the focus too.

    set oldFocus [focus]
    set oldGrab [grab current $w]
    if {$oldGrab != ""} {
	set grabStatus [grab status $oldGrab]
    }
    grab $w
    focus $data(ent)

    # Wait for the user to respond, then restore the focus and
    # return the index of the selected button.  Restore the focus
    # before deleting the window, since otherwise the window manager
    # may take the focus away so we can't redirect it.  Finally,
    # restore any grab that was in effect.

    tkwait variable ${dataName}(selectedOption)
    catch {focus $oldFocus}
    grab release $w
    wm withdraw $w
    if {$oldGrab != ""} {
	if {$grabStatus == "global"} {
	    grab -global $oldGrab
	} else {
	    grab $oldGrab
	}
    }

    return [::list $tkPriv(selectOption_status) $tkPriv(selectOption)]
}

# tkPOption_Config --
#
#	Configures the TK filedialog according to the argument list
#
proc tkPOption_Config {dataName argList} {
    upvar #0 $dataName data

    # 1: the configuration specs
    #
    set specs {
	{-options "" "" "" }
	{-parent "" "" "."}
	{-title "" "" "Printer Selection"}
	{-withsave "" "" 1}
    }

    # 2: default values 
    #
    set tkPriv(selectOption) data(-options)

    # 3: parse the arguments
    #
    tclParseConfigSpec $dataName $specs "" $argList

    if {![string compare $data(-title) ""]} {
	set data(-title) "Printer Options"
    }

    if {![winfo exists $data(-parent)]} {
	error "bad window path name \"$data(-parent)\""
    }
}

proc tkPOption_Create {w} {
    set dataName [lindex [split $w .] end]
    upvar #0 $dataName data
    global tk_library
    global tkPriv

    toplevel $w -class TkPOption

    # f2: the frame with the the "printer options" field
    #
    set f2 [frame $w.f2 -bd 0]
    label $f2.lab -text "Printer options:" -anchor w -width 14 -under 5 -pady 0 -padx 5
    set data(ent) [entry $f2.ent -textvariable tkPriv(selectOption)]

    # f3: the frame with the OK and cancel buttons
    #
    set f3 [frame $w.f3 -bd 0]

    set data(okBtn)     [button $f3.ok     -text OK     -under 0 -width 6 \
	-default active -pady 3]
    set data(cancelBtn) [button $f3.cancel -text Cancel -under 0 -width 6\
	-default normal -pady 3]
    if { $data(-withsave) == 0 } {
      ;# Can't put up save button
    } else {
      set data(saveBtn) [button $f3.save -text "Save" -under 0 -width 6\
        -default normal -pady 3 -command "tkPOption_SaveCmd $w" ]
      bind $w <Alt-s> "tkButtonInvoke $data(saveBtn)"
    }

    # pack the widgets in f2 and f3
    #
    pack $data(okBtn) -side left -padx 4 -anchor center
    pack $f2.lab -side left -padx 4
    pack $f2.ent -expand yes -fill x -padx 4 -pady 0
    
    pack $data(cancelBtn) -side right -padx 4 -anchor center

    if [ info exist data(saveBtn) ] {
      pack $data(saveBtn) -side left -padx 4 -anchor center -expand 1
    }

    # Pack all the frames together. We are done with widget construction.
    #
    pack $f3 -side bottom -fill x
    pack $f2 -side bottom -fill x

    # Set up the event handlers
    #
    bind $data(ent) <Return>  "tkPOption_OkCmd $w"
    
    $data(okBtn)     config -command "tkPOption_OkCmd $w"
    $data(cancelBtn) config -command "tkPOption_CancelCmd $w"

    bind $w <Alt-n> "focus $data(ent)"
    bind $w <KeyPress-Escape> "tkButtonInvoke $data(cancelBtn)"
    bind $w <Alt-c> "tkButtonInvoke $data(cancelBtn)"

    wm protocol $w WM_DELETE_WINDOW "tkPOption_CancelCmd $w"

    # Build the focus group for all the entries
    #
    tkFocusGroup_Create $w
}

# Gets called when user presses the "OK" button
#
proc tkPOption_OkCmd {w} {
    global tkPriv
    upvar #0 [winfo name $w] data

    set text $tkPriv(selectOption)
    set data(selectedOption) $text
    set tkPriv(selectOption_status) 1
    return $tkPriv(selectOption)
}

# Gets called when user presses the "Save" button
# Options saved and then treated like OK.
# Perhaps the save script itself should be something passed in to the creation of
# the command itself
#
proc tkPOption_SaveCmd {w} {
    global tkPriv
    upvar #0 [winfo name $w] data

    set text $tkPriv(selectOption)
    set data(selectedOption) $text
    set tkPriv(selectOption_status) 2
    return $tkPriv(selectOption)
}

# Gets called when user presses the "Cancel" button
#
proc tkPOption_CancelCmd {w} {
    upvar #0 [winfo name $w] data
    global tkPriv

    set data(selectedOption) ""
    set tkPriv(selectOption_status) 0 
    return $tkPriv(selectOption)
}

# tkPOption_Done --
#
#	Gets called when user has input a valid filename.  Pops up a
#	dialog box to confirm selection when necessary. Sets the
#	tkPriv(selectOption) variable, which will break the "tkwait"
#	loop in tkPOption and return the selected filename to the
#	script that calls tk_getOpenFile or tk_getSaveFile
#
proc tkPOption_Done {w {selectOption ""}} {
    upvar #0 [winfo name $w] data
    global tkPriv

    set data(selectedOption) "$selectOption"
    set tkPriv(selectOption_status) 1
    return $tkPriv(selectOption)
}
